;; Format
;; #XXXXCCCC
;; # - '#' character
;; XXXX - 16-bit value as hex ASCII string (for example, 7FFF for 32767)
;; CCCC - CRC-16 value as hex ASCII string (for example, 7FFF for 32767)

IE      .EQU    0A8H
SCON    .EQU    098H
SBUF    .EQU    099H
TMOD    .EQU    089H
TL1     .EQU    08BH
TH1     .EQU    08DH
ET1     .EQU    0ABH
TR1     .EQU    08EH
PCON    .EQU    087H

DPL     .EQU    082H
DPH     .EQU    083H

TI      .EQU    099H
RI      .EQU    098H

P0      .EQU    80H
P1      .EQU    90H
P2      .EQU    0A0H
P3      .EQU    0B0H

P3.0    .EQU    0B0H
P3.1    .EQU    0B1H
P3.2    .EQU    0B2H
P3.3    .EQU    0B3H
P3.4    .EQU    0B4H
P3.5    .EQU    0B5H
P3.6    .EQU    0B6H
P3.7    .EQU    0B7H

B       .EQU    0F0H
ACC     .EQU    0E0H

B.7     .EQU    0F7H
B.6     .EQU    0F6H
B.5     .EQU    0F5H
B.4     .EQU    0F4H
B.3     .EQU    0F3H
B.2     .EQU    0F2H
B.1     .EQU    0F1H
B.0     .EQU    0F0H
ACC.7   .EQU    0E7H
ACC.6   .EQU    0E6H
ACC.5   .EQU    0E5H
ACC.4   .EQU    0E4H
ACC.3   .EQU    0E3H
ACC.2   .EQU    0E2H
ACC.1   .EQU    0E1H
ACC.0   .EQU    0E0H

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
REL96   .EQU    0FDH    ; UART 9600 bits/second (for 11.059 MHz oscillator)

CH_START .EQU   '#'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.ORG    0
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Program entry point
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        ; ந : 9600 /, 8 /ᨬ, 1 ⮯ ,  ⭮
        MOV     IE, #0
        MOV     TMOD, #20H
        MOV     TH1, #REL96
        MOV     TL1, #REL96
        ANL     PCON, #7FH
        CLR     ET1
        MOV     SCON, #50H
        SETB    TR1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main program loop
;; Generates a triangular wave: 0, 1000, ..., 65000, 64000, 63000, ..., 0, ...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        ; R5R4 = 0:
        MOV     R5, #0
        MOV     R4, #0
        ; R3R2 = 1000:
        MOV     R3, #1000 >> 8
        MOV     R2, #1000 & 255
MAIN:
        ACALL   DELAY
        ACALL   SEND

        ; if R5R4 = 65000, then R3R2 = -1000:
        CLR     C
        MOV     A, R4
        SUBB    A, #65000 & 255
        MOV     B, A
        MOV     A, R5
        SUBB    A, #65000 >> 8
        ORL     A, B
        JNZ     L1

        MOV     R3, #(-1000) >> 8
        MOV     R2, #(-1000) & 255

L1:
        ; if R5R4 = 0, then R3R2 = 1000:
        CLR     C
        MOV     A, R4
        SUBB    A, #0 & 255
        MOV     B, A
        MOV     A, R5
        SUBB    A, #0 >> 8
        ORL     A, B
        JNZ     L2

        MOV     R3, #1000 >> 8
        MOV     R2, #1000 & 255

L2:
        ; R5R4 = R5R4 + R3R2:
        MOV     A, R4
        ADD     A, R2
        MOV     R4, A
        MOV     A, R5
        ADDC    A, R3
        MOV     R5, A

        SJMP    MAIN

; 200 ms delay:
DELAY:
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        ACALL   DEL20
        RET

DEL20:
; 2+1+144*(2+62*2+2)+2=18437 clocks (11.095 MHz / 12 * 20 ms = 18432 clocks)
        MOV     R1,#144
LPEX:
        MOV     R0,#62
        NOP
        DJNZ    R0,$
        DJNZ    R1,LPEX
        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Send 16-bit value and its CRC-16 to UART
;; IN:  R5R4 - 16-bit value to be sent over UART
;; OUT: none
;; USED REGS: A, R6, R7, DPTR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
SEND:
        ; send start character:
        MOV     A, #CH_START
        ACALL   PUTCH

        ; send 16-bit value
        MOV     A, R5           ; MSB first 
        SWAP    A
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R5 
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R4           ; LSB second
        SWAP    A
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R4 
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH

        ; calculate CRC-16 of the 16-bit value:
        ACALL   CRC16_START
        MOV     A, R5           ; MSB first 
        ACALL   CRC16_CALC_BYTE
        MOV     A, R4           ; LSB second
        ACALL   CRC16_CALC_BYTE
        ACALL   CRC16_FINISH    ; R7R6 = CRC-16 of R5R4

        ; send CRC-16:
        MOV     A, R7           ; MSB first 
        SWAP    A
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R7 
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R6           ; LSB second
        SWAP    A
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH
        MOV     A, R6 
        ANL     A, #0FH
        ACALL   HEXDIGIT2CHAR
        ACALL   PUTCH

        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ਥ ᨬ   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;
GETCH:  JNB     RI, $
        MOV     A, SBUF
        CLR     RI
        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; । ᨬ   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PUTCH:  MOV     SBUF, A
        JNB     TI, $
        CLR     TI
        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Converts character '0'...'9','A'...'F' to 4-bit value
;; IN:  A = character
;; OUT: A = 4-bit value (iff no carry, junk otherwise)
;; USED REGS: none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CHAR2HEXDIGIT:
        CLR     C
        SUBB    A, #'9'+1
        JNC     CHAR2HEXDIGIT_1 ; ᫨ > "9", ஢  "A"..."F"
        ADD     A, #'9'+1
        CLR     C
        SUBB    A, #'0'         ; CY=1 (訡), ᫨ < "0",
                                ;  ᨬ  "0"  "9",   0  9
        RET
CHAR2HEXDIGIT_1:
        ADD     A, #'9'+1
        CLR     C
        SUBB    A, #'F'+1
        CPL     C
        JNC     CHAR2HEXDIGIT_2 ; CY=1 (訡), ᫨ > "F"
        RET
CHAR2HEXDIGIT_2:
        ADD     A, #'F'+1
        CLR     C
        SUBB    A, #'A'
        JNC     CHAR2HEXDIGIT_3 ; CY=1 (訡), ᫨ < "A"
        RET
CHAR2HEXDIGIT_3:
        ADD     A, #0Ah         ; ᨬ  "A"  "F",   0Ah  0Fh
        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Converts 4-bit value to character '0'...'9','A'...'F'
;; IN:  A = 4-bit value
;; OUT: A = character
;; USED REGS: none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
HEXDIGIT2CHAR:
        INC     A
        MOVC    A, @A+PC
        RET
HEX_TAB .TEXT   "0123456789ABCDEF"

CRC16_POLY .EQU  08005h ; X**16 + X**15 + X**2 + X**0
CRC16_INIT .EQU  00000h

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initializes CRC16
;; IN:  none
;; OUT: R7R6 - CRC16 initial value
;; USED REGS: none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CRC16_START:
        MOV     R7, #(CRC16_INIT >> 8)
        MOV     R6, #(CRC16_INIT & 255)
        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Calculates/updates CRC16 value for a new data byte
;; IN:  A - new byte
;;      R7R6 - CRC16 value (initial or intermediate)
;; OUT: R7R6 - CRC16 updated value
;; USED REGS: A, DPTR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CRC16_CALC_BYTE:
        XCH     A, R6   ; A=CRCLO, R6=new byte
        XCH     A, R7   ; R7=CRCLO, A=CRCHI

        CLR     C
        RLC     A
        MOV     DPTR, #CRC16_TAB
        JNC     CRC16_CALC_BYTE_1
        INC     DPH
CRC16_CALC_BYTE_1:
        ADD     A, DPL
        MOV     DPL, A
        JNC     CRC16_CALC_BYTE_2
        INC     DPH
CRC16_CALC_BYTE_2:

        CLR     A
        MOVC    A, @A+DPTR
        XRL     A, R6
        MOV     R6, A
        INC     DPTR
        CLR     A
        MOVC    A, @A+DPTR
        XRL     A, R7
        MOV     R7, A

        RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Finishes CRC16 calculation (must be called after last data byte)
;; IN:  R7R6 - CRC16 value (after last data byte processed
;;             by CRC16_CALC_BYTE)
;; OUT: R7R6 - CRC16 final value
;; USED REGS: A, DPTR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CRC16_FINISH:
        CLR     A
        ACALL   CRC16_CALC_BYTE
        CLR     A
        AJMP    CRC16_CALC_BYTE

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CRC16 table for CRC16_POLY = 08005h (X**16 + X**15 + X**2 + X**0)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CRC16_TAB:
        .DW 00000h,08005h,0800Fh,0000Ah,0801Bh,0001Eh,00014h,08011h
        .DW 08033h,00036h,0003Ch,08039h,00028h,0802Dh,08027h,00022h
        .DW 08063h,00066h,0006Ch,08069h,00078h,0807Dh,08077h,00072h
        .DW 00050h,08055h,0805Fh,0005Ah,0804Bh,0004Eh,00044h,08041h
        .DW 080C3h,000C6h,000CCh,080C9h,000D8h,080DDh,080D7h,000D2h
        .DW 000F0h,080F5h,080FFh,000FAh,080EBh,000EEh,000E4h,080E1h
        .DW 000A0h,080A5h,080AFh,000AAh,080BBh,000BEh,000B4h,080B1h
        .DW 08093h,00096h,0009Ch,08099h,00088h,0808Dh,08087h,00082h
        .DW 08183h,00186h,0018Ch,08189h,00198h,0819Dh,08197h,00192h
        .DW 001B0h,081B5h,081BFh,001BAh,081ABh,001AEh,001A4h,081A1h
        .DW 001E0h,081E5h,081EFh,001EAh,081FBh,001FEh,001F4h,081F1h
        .DW 081D3h,001D6h,001DCh,081D9h,001C8h,081CDh,081C7h,001C2h
        .DW 00140h,08145h,0814Fh,0014Ah,0815Bh,0015Eh,00154h,08151h
        .DW 08173h,00176h,0017Ch,08179h,00168h,0816Dh,08167h,00162h
        .DW 08123h,00126h,0012Ch,08129h,00138h,0813Dh,08137h,00132h
        .DW 00110h,08115h,0811Fh,0011Ah,0810Bh,0010Eh,00104h,08101h
        .DW 08303h,00306h,0030Ch,08309h,00318h,0831Dh,08317h,00312h
        .DW 00330h,08335h,0833Fh,0033Ah,0832Bh,0032Eh,00324h,08321h
        .DW 00360h,08365h,0836Fh,0036Ah,0837Bh,0037Eh,00374h,08371h
        .DW 08353h,00356h,0035Ch,08359h,00348h,0834Dh,08347h,00342h
        .DW 003C0h,083C5h,083CFh,003CAh,083DBh,003DEh,003D4h,083D1h
        .DW 083F3h,003F6h,003FCh,083F9h,003E8h,083EDh,083E7h,003E2h
        .DW 083A3h,003A6h,003ACh,083A9h,003B8h,083BDh,083B7h,003B2h
        .DW 00390h,08395h,0839Fh,0039Ah,0838Bh,0038Eh,00384h,08381h
        .DW 00280h,08285h,0828Fh,0028Ah,0829Bh,0029Eh,00294h,08291h
        .DW 082B3h,002B6h,002BCh,082B9h,002A8h,082ADh,082A7h,002A2h
        .DW 082E3h,002E6h,002ECh,082E9h,002F8h,082FDh,082F7h,002F2h
        .DW 002D0h,082D5h,082DFh,002DAh,082CBh,002CEh,002C4h,082C1h
        .DW 08243h,00246h,0024Ch,08249h,00258h,0825Dh,08257h,00252h
        .DW 00270h,08275h,0827Fh,0027Ah,0826Bh,0026Eh,00264h,08261h
        .DW 00220h,08225h,0822Fh,0022Ah,0823Bh,0023Eh,00234h,08231h
        .DW 08213h,00216h,0021Ch,08219h,00208h,0820Dh,08207h,00202h

        .END
