#include "StdAfx.h"
#include "Buffer.h"

CBuffer::CBuffer (size_t Size) : NotEmpty (FALSE, FALSE)
{
  Size++;
  pBuf = new unsigned char[Size];
  BufSize = Size;
  IndexHead = IndexTail = 0;
}

CBuffer::~CBuffer()
{
  delete [] pBuf;
}

void CBuffer::Reset()
{
  CSingleLock lock (&Lock, TRUE);
  IndexHead = IndexTail = 0;
}

int CBuffer::PutChar (unsigned char Char)
{
  unsigned int i;
  CSingleLock lock (&Lock, TRUE);

  i = IndexHead + 1;
  if (i >= BufSize)
    i = 0;

  if (i == IndexTail)
    return 0;

  pBuf[IndexHead] = Char;
  IndexHead = i;

  // Flag the event!
  lock.Unlock();
  NotEmpty.SetEvent();
  
  return 1;
}

int CBuffer::GetChar (unsigned char* pChar)
{
  CSingleLock lock (&Lock, TRUE);

  if (IndexHead == IndexTail)
    return 0;

  *pChar = pBuf[IndexTail++];

  if (IndexTail >= BufSize)
    IndexTail = 0;

  // Keep the event valid if need be!
  if (IndexHead != IndexTail)
  {
    lock.Unlock();
    NotEmpty.SetEvent();
  }
  else
  {
    lock.Unlock();
    NotEmpty.ResetEvent();
  }

  return 1;
}

void CBuffer::Clear()
{
  CSingleLock lock (&Lock, TRUE);

  IndexTail = IndexHead;

  lock.Unlock();
  NotEmpty.ResetEvent();
}
