#include "CRC16.h"

void CRC16_StartCalc   (uint16* pCRC16Value)
{
  *pCRC16Value = CRC16_INIT;
}

void CRC16_Calc        (uint16* pCRC16Value, const uint8* pDataBytes,
                        size_t DataBytesCount)
{
  uint16 CRC16 = *pCRC16Value;
  while (DataBytesCount--)
  {
    int i;
    uint8 Byte = (*pDataBytes++) & 0xFF;
    for (i=0; i<8; i++)
    {
      int f = (CRC16 & 0x8000U) != 0;
      CRC16 <<= 1;
      if (Byte & 0x80)
        CRC16 |= 1;
      Byte <<= 1;
      if (f)
        CRC16 ^= CRC16_POLY;
    }
  }
  *pCRC16Value = CRC16 & 0xFFFFU;
}

void CRC16_InitTable   (uint16 aCRC16Tab[0x100])
{
  int i;

  for (i=0; i<0x100; i++)
  {
    uint8 b=0;
    aCRC16Tab[i] = ((uint16)i)<<8;
    CRC16_Calc (&aCRC16Tab[i], &b, 1);
  }
}

void CRC16_CalcByTable (const uint16 aCRC16Tab[0x100], uint16* pCRC16Value,
                        const uint8* pDataBytes, size_t DataBytesCount)
{
  uint16 CRC16 = *pCRC16Value;
  while (DataBytesCount--)
  {
    uint8 i = (CRC16 >> 8) & 0xFF;
    CRC16 = (CRC16 << 8) | ((*pDataBytes++) & 0xFF);
    CRC16 ^= aCRC16Tab[i];
  }
  *pCRC16Value = CRC16 & 0xFFFFU;
}

void CRC16_FinishCalc  (uint16* pCRC16Value)
{
  uint8 b = 0;
  CRC16_Calc (pCRC16Value, &b, 1);
  CRC16_Calc (pCRC16Value, &b, 1);
}

#if 0
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define DATA_LEN 11

int main()
{
  uint8 aData[DATA_LEN+2];
  int i;
  time_t t;
  uint16 CRC16;
  uint16 aCRC16Table[0x100];

  CRC16_InitTable (aCRC16Table);
  srand ((unsigned)time(&t));
  printf ("%d 8-bit random data bytes:\n", (int)DATA_LEN);
  for (i=0; i<DATA_LEN; i++)
  {
    aData[i] = rand() & 0xFF;
    printf ("0x%02X ", (unsigned)aData[i]);
  }
  printf("\n");

  CRC16_StartCalc (&CRC16);
  CRC16_Calc (&CRC16, aData, DATA_LEN);
  CRC16_FinishCalc (&CRC16);
  printf ("CRC16slow=0x%04X\n", (unsigned)CRC16);

  CRC16_StartCalc (&CRC16);
  CRC16_CalcByTable (aCRC16Table, &CRC16, aData, DATA_LEN);
  CRC16_FinishCalc (&CRC16);
  printf ("CRC16fast=0x%04X\n", (unsigned)CRC16);

  aData[DATA_LEN] = (CRC16 >> 8) & 0xFF;
  aData[DATA_LEN+1] = CRC16 & 0xFF;
  CRC16_StartCalc (&CRC16);
  CRC16_CalcByTable (aCRC16Table, &CRC16, aData, DATA_LEN+2);
  CRC16_FinishCalc (&CRC16);
  printf ("Now, CRC16 of the same data and its CRC16 appended to it, must be 0:\n", (unsigned)CRC16);
  printf ("CRC16=0x%04X\n", (unsigned)CRC16);

  return 0;
}
#endif
